/*
 * A simple flash mapping code for BCM963xx board flash memory
 * It is simple because it only treats all the flash memory as ROM
 * It is used with chips/map_rom.c
 *
 *  Song Wang (songw@broadcom.com)
 */
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/config.h>
#include <linux/mtd/partitions.h>

#include <board.h>
#include <bcmTag.h>
#define  VERSION	"1.0"


extern PFILE_TAG kerSysImageTagGet(void);

static struct mtd_info *mymtd;
static struct mtd_partition brcm_partition_info[5];//JIM increase from 4 to 5 for POT

#define CFE_ADDR      	0xbfc00000
#define CFE_SIZE      	64 << 10
#define FLASH_2M_SIZE 	2048 << 10
#define FLASH_4M_SIZE 	4096 << 10
#define NVRAM_SIZE    	64 << 10
#define TAG_SIZE	0x100
#define FS_KERNEL_SIZE_4M	0x3E0000
#define NVRAM_ADDR_4M	0x3F0000

static map_word brcm_physmap_read16(struct map_info *map, unsigned long ofs)
{
	map_word val;
	
	val.x[0] = __raw_readw(map->map_priv_1 + ofs);
	
	return val;
}

void brcm_physmap_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
	memcpy_fromio(to, map->map_priv_1 + from, len);
}

static void brcm_physmap_write16(struct map_info *map, __u16 d, unsigned long adr)
{
	__raw_writew(d, map->map_priv_1 + adr);
	mb();
}

void brcm_physmap_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
	memcpy_toio(map->map_priv_1 + to, from, len);
}

struct map_info brcm_physmap_map = {
	.name = "Physically mapped flash",
	.bankwidth	= 2,
	.read = brcm_physmap_read16,
	.copy_from = brcm_physmap_copy_from,
	.write = brcm_physmap_write16,
	.copy_to = brcm_physmap_copy_to

};



int __init init_brcm_physmap(void)
{
        PFILE_TAG pTag = NULL;
        u_int32_t rootfs_addr, kernel_addr,fs_len,cfe_len;
        FLASH_ADDR_INFO info;

        kerSysFlashAddrInfoGet( &info );

        /* Read the flash memory map from flash memory. */
        if (!(pTag = kerSysImageTagGet())) {
                printk("Failed to read image tag from flash\n");
                return -EIO;
        }

        rootfs_addr = (u_int32_t) simple_strtoul(pTag->rootfsAddress, NULL, 10);
        kernel_addr = (u_int32_t) simple_strtoul(pTag->kernelAddress, NULL, 10);
	
	
	brcm_physmap_map.size = FLASH_4M_SIZE;
	
	fs_len = kernel_addr - rootfs_addr;
	cfe_len=CFE_SIZE;
	/* Ron mapping from fs */
	brcm_physmap_map.map_priv_1 = (unsigned long)CFE_ADDR;

	if (!brcm_physmap_map.map_priv_1) {
		printk("Wrong flash starting address\n");
		return -EIO;
	}
	
	if (brcm_physmap_map.size <= 0) {
		printk("Wrong flash size\n");
		return -EIO;
	}	
	
	mymtd = do_map_probe("cfi_probe", &brcm_physmap_map);
	if (!mymtd)
		return -EIO;
	
	mymtd->owner = THIS_MODULE;
	/* Ron file system */
	brcm_partition_info[0].name = "fs";
	brcm_partition_info[0].offset = (cfe_len + TAG_SIZE);
	brcm_partition_info[0].size = fs_len;
	brcm_partition_info[0].mask_flags = 0;
	
	/* Ron tag + file system + kernel */
	brcm_partition_info[1].name = "tag+fs+kernel";
	brcm_partition_info[1].offset = cfe_len;
	brcm_partition_info[1].size = FS_KERNEL_SIZE_4M; 
	brcm_partition_info[1].mask_flags = 0;

	/* Ron bootloader */
	brcm_partition_info[2].name = "bootloader";
	brcm_partition_info[2].offset = 0x00;
	brcm_partition_info[2].size = cfe_len;
	brcm_partition_info[2].mask_flags = 0;

	/* Ron nvram */
	brcm_partition_info[3].name = "nvram";
	brcm_partition_info[3].offset = NVRAM_ADDR_4M;
	brcm_partition_info[3].size = NVRAM_SIZE;
	brcm_partition_info[3].mask_flags = 0;
	
	/* JIM add for mapping to bootloader */
	brcm_partition_info[4].name = "bootloader";
	brcm_partition_info[4].offset = 0x00;
	brcm_partition_info[4].size = cfe_len;
	brcm_partition_info[4].mask_flags = 0;
	
	add_mtd_partitions(mymtd, brcm_partition_info, 5);//JIM increase from 4 to 5 for POT

	return 0;

}

static void __exit cleanup_brcm_physmap(void)
{
if (mymtd) {
	del_mtd_partitions(mymtd);
	del_mtd_device(mymtd);
	map_destroy(mymtd);
}
if (brcm_physmap_map.map_priv_1) {
	brcm_physmap_map.map_priv_1 = 0;
}
}

module_init(init_brcm_physmap);
module_exit(cleanup_brcm_physmap);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Song Wang songw@broadcom.com");
MODULE_DESCRIPTION("Configurable MTD map driver for read-only root file system");
